!+
! Module: coordinate_old
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains several routines to do with coordinate transformations.
! 
! Comments: TODO:
! 
! $Id: coordinate.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module coordinate_old

  use dal
  use cal	! contains 'use caloaldefs' and 'use caltypes'
  use oal	! contains 'use oaltypes'. For OAL_toAttitudeMatrix.
  use errorHandling

  use constants
  use wcs_aux		! for WcsT in getThetaPhiMaps().
  use array_utils	! for vectorCross
  use cal_aux 		! for getInstrumentAttitude.
!  use dump		! for printMatrix

  implicit none 

  private 

  public raDecToInst
  public instToRaDec
  public raDecToTan
  public tanToRaDec
  public instToTan_old
  public tanToInst
  public getThetaPhiMaps

  interface raDecToInst
    module procedure raDecToInstScalar
    module procedure raDecToInstVector
  end interface

  interface instToRaDec
    module procedure instToRaDecScalar
    module procedure instToRaDecVector
  end interface

  interface raDecToTan
    module procedure raDecToTanScalar
    module procedure raDecToTanVector
  end interface

  interface tanToRaDec
    module procedure tanToRaDecScalar
    module procedure tanToRaDecVector
  end interface

  interface instToTan_old
    module procedure instToTanScalar, instToTanVector, instToTanArray
  end interface

  interface tanToInst
    module procedure tanToInstScalar, tanToInstVector, tanToInstArray
  end interface

  real(single), public, parameter :: detUnitArcsec = 0.05 ! this should be centrally accessible somewhere? (But where?? In a parameter file?)

  type, public :: PointType
    real(single) :: x, y
  end type

contains 

  !---------------------------------------------------------------------
  subroutine raDecToInstScalar(raDeg, decDeg, scAttitude&
  , timeStamp, xDet, yDet, thetaArcsec, phi, status)
  !---------------------------------------------------------------------
! was raDecToDetXY
!*** note! change raDeg-> ra and decDeg->dec.

  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in) :: raDeg,&
                                                decDeg
    type(SpacecraftAttitudeType), intent(in) :: scAttitude
    ! typedef from caloalutils/src/caloaldefs.f90
    real(double),                 intent(in) :: timeStamp
    real(single)     , optional, intent(out) :: xDet,&
                                                yDet
    real(double)     , optional, intent(out) :: thetaArcsec,&
                                                phi
    integer,           optional, intent(out) :: status


    real(double) :: raDegVector(1), decDegVector(1)
    real(single) :: xDetVector(1), yDetVector(1)
    real(double) :: thetaArcsecVector(1), phiVector(1)
    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start

    if (present(xDet)) xDet = 0.0
    if (present(yDet)) yDet = 0.0
    if (present(thetaArcsec)) thetaArcsec = 0.0d0
    if (present(phi)) phi = 0.0d0

    raDegVector(1)  = raDeg
    decDegVector(1) = decDeg

    if (present(xDet) .and. present(yDet)) then
      call raDecToInstVector(raDegVector, decDegVector, scAttitude&
        , timeStamp, xDetVector, yDetVector, status=foreignStatus)

      xDet = xDetVector(1)
      yDet = yDetVector(1)

    else if (present(thetaArcsec) .and. present(phi)) then
      call raDecToInstVector(raDegVector, decDegVector, scAttitude&
        , timeStamp, thetaArcsec=thetaArcsecVector, phi=phiVector&
        , status=foreignStatus)

      thetaArcsec = thetaArcsecVector(1)
      phi         = phiVector(1)

    else
      if (present(status)) status = -1
  return
    end if

    if (present(status)) status = foreignStatus

  end subroutine raDecToInstScalar

  !---------------------------------------------------------------------
  subroutine raDecToInstVector(raDeg, decDeg, scAttitude&
  , timeStamp, xDet, yDet, thetaArcsec, phi, status)
  !---------------------------------------------------------------------
! was raDecToDetXY
!*** note! change raDeg-> ra and decDeg->dec.

  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in) :: raDeg(:),&
                                                decDeg(size(raDeg))
    type(SpacecraftAttitudeType), intent(in) :: scAttitude
    ! typedef from caloalutils/src/caloaldefs.f90
    real(double),                 intent(in) :: timeStamp
    real(single)     , optional, intent(out) :: xDet(size(raDeg)),&
                                                yDet(size(raDeg))
    real(double)     , optional, intent(out) :: thetaArcsec(size(raDeg)),&
                                                phi(size(raDeg))
    integer,           optional, intent(out) :: status

    type(BoresightT) ::&
      boresight ! typedef from cal/f90/caltypes.f90
    type(EquatorialDirectionType), allocatable ::&
      skyVectors(:) ! typedef from caloalutils/src/caloaldefs.f90
    real(double), dimension(:), pointer ::&
      thetaArcsecArg, phiArg, cc2x, cc2y, cc2z
    real(double) ::&
      mmToDetUnits, focalLengthMm
    integer(int32) ::&
      numPositions, i
    integer ::&
      localStatus
!logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Noisy, 'ssclib: ra/dec to XY.')

    localStatus = 0

    if (present(xDet)) xDet = 0.0
    if (present(yDet)) yDet = 0.0
    if (present(thetaArcsec)) thetaArcsec = 0.0d0
    if (present(phi)) phi = 0.0d0

    numPositions = size(raDeg)

!    call CAL_setState(instrument=instrumentId, randomize=.false.&
!    , ccdModeId=PRIME_FULL_WINDOW, accuracyLevel=ACCURACY_LOW)
!!*** Need the ccdModeId??

    boresight = CAL_getBoresight(timeStamp) ! needs instrument
    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm) ! needs instrument
    mmToDetUnits = RadToArcsec / focalLengthMm / detUnitArcsec

    allocate(skyVectors(numPositions))
    do i = 1, numPositions
      skyVectors(i)%ra = raDeg(i) * PiOn180
      do while(skyVectors(i)%ra .ge. TwoPi)
        skyVectors(i)%ra = skyVectors(i)%ra - TwoPi
      end do
      do while(skyVectors(i)%ra .lt. 0.0)
        skyVectors(i)%ra = skyVectors(i)%ra + TwoPi
      end do

      if (decDeg(i) .lt. -90.0 .or. decDeg(i) .gt. 90.0) then
        if (present(status)) status = -2
  return
      end if

      skyVectors(i)%dec = min(max(decDeg(i) * PiOn180, -PiOverTwo)&
      , PiOverTwo)
!print *, 'i, skyVectors(i)%ra, skyVectors(i)%dec: ', i, skyVectors(i)%ra, skyVectors(i)%dec
    end do

    call CAL_withSacCoordOrientation(boresight, scAttitude) ! needs instrument
    call CAL_projectOntoTelCoord(boresight, skyVectors, thetaArcsecArg, phiArg) ! needs instrument.
    ! NOTE! Units of phi are radians.
!print *, 'thetaArcsecArg(0), phiArg(0): ', thetaArcsecArg(0), phiArg(0)
    call CAL_telCoordToCamCoord2(thetaArcsecArg, phiArg, cc2x, cc2y, cc2z) ! needs only instrument?

    if (present(xDet)) xDet = cc2X * mmToDetUnits
    if (present(yDet)) yDet = cc2Y * mmToDetUnits
    if (present(thetaArcsec)) thetaArcsec = thetaArcsecArg
    if (present(phi)) phi = phiArg

    call CAL_releaseMemory(thetaArcsecArg)
    call CAL_releaseMemory(phiArg)
    call CAL_releaseMemory(cc2x, cc2y, cc2z)

    deallocate(skyVectors)

    if (present(status)) status = localStatus

  end subroutine raDecToInstVector

  !---------------------------------------------------------------------
  subroutine instToRaDecScalar(raDeg, decDeg, scAttitude&
  , timeStamp, xDet, yDet, thetaArcsec, phi, status)
  !---------------------------------------------------------------------
! was detXYtoRaDecScalar
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(out) ::&
      raDeg,&
      decDeg
    type(SpacecraftAttitudeType), intent(in)  ::&
      scAttitude ! typedef from caloalutils/src/caloaldefs.f90
    real(double),                 intent(in)  ::&
      timeStamp
    real(single),       optional, intent(in)  ::&
      xDet,&
      yDet
    real(double),       optional, intent(in)  ::&
      thetaArcsec,&
      phi
    integer,            optional, intent(out) ::&
      status

    real(single) ::&
      vectorXdet(1),&
      vectorYdet(1)
    real(double) ::&
      vectorRaDeg(1),&
      vectorDecDeg(1),&
      vectorThetaArcsec(1),&
      vectorPhi(1)
    integer ::&
      foreignStatus

    !-------------------------------------------------------------------
    ! Start

    raDeg  = 0.0d0 ! default
    decDeg = 0.0d0 ! default

    if (present(xDet) .and. present(yDet)) then
      vectorXdet(1) = xDet
      vectorYdet(1) = yDet

      call instToRaDecVector(vectorRaDeg, vectorDecDeg, scAttitude&
        , timeStamp, vectorXdet, vectorYdet, status=foreignStatus)

    else if (present(thetaArcsec) .and. present(phi)) then
      vectorThetaArcsec(1) = thetaArcsec
      vectorPhi(1)         = phi

      call instToRaDecVector(vectorRaDeg, vectorDecDeg, scAttitude&
        , timeStamp, thetaArcsec=vectorThetaArcsec, phi=vectorPhi&
        , status=foreignStatus)

    else
      if (present(status)) status = -1
  return
    end if

    raDeg  = vectorRaDeg(1)
    decDeg = vectorDecDeg(1)
    if (present(status)) status = foreignStatus

  end subroutine instToRaDecScalar

  !---------------------------------------------------------------------
  subroutine instToRaDecVector(raDeg, decDeg, scAttitude&
  , timeStamp, xDet, yDet, thetaArcsec, phi, status)
  !---------------------------------------------------------------------
! was detXYtoRaDecVector
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(out) :: raDeg(:),&
                                                 decDeg(size(raDeg))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    ! typedef from caloalutils/src/caloaldefs.f90
    real(double),                 intent(in)  :: timeStamp
    real(single),       optional, intent(in)  :: xDet(size(raDeg)),&
                                                 yDet(size(raDeg))
    real(double),       optional, intent(in)  :: thetaArcsec(size(raDeg)),&
                                                 phi(size(raDeg))
    integer,            optional, intent(out) :: status

    real(double), dimension(:), pointer ::&
      localThetaArcsec, localPhi, cc2x, cc2y, cc2z
    real(double) ::&
      mmToDetUnits, focalLengthMm, tanTheta, detCartesian(3),&
      skyCartesian(3), transposedInstrumAttitudeMatrix(3,3), ra, dec
    integer(int32) ::&
      numPositions, i
    integer ::&
      localStatus
    type(SpacecraftAttitudeType) ::&
      instrumAttitude

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    raDeg  = 0.0d0 ! default
    decDeg = 0.0d0 ! default

    numPositions = size(raDeg)

    if (present(xDet) .and. present(yDet)) then
      call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
      mmToDetUnits = RadToArcsec / focalLengthMm / detUnitArcsec

      allocate(cc2X(numPositions), cc2Y(numPositions))
      cc2X = xDet / mmToDetUnits
      cc2Y = yDet / mmToDetUnits
      cc2z = focalLengthMm

      call CAL_camCoord2ToTelCoord(cc2x, cc2y, cc2z, localThetaArcsec, localPhi)

      deallocate(cc2X, cc2Y, cc2z)

    else if (present(thetaArcsec) .and. present(phi)) then
      allocate(localThetaArcsec(numPositions), localPhi(numPositions))
      localThetaArcsec = thetaArcsec
      localPhi = phi

!print *, 'localThetaArcsec(1):', localThetaArcsec(1)
!print *, '        localPhi(1):', localPhi(1)

    else
      if (present(status)) status = -1
  return
    end if

    instrumAttitude = getInstrumentAttitude(scAttitude, timeStamp)

    call OAL_toAttitudeMatrix(instrumAttitude, transposedInstrumAttitudeMatrix)
    !
    ! The array returned by this call represents the instrument attitude
    ! matrix AIN_sky which has, in its ith row, a vector of direction
    ! cosines in the sky basis of the ith axis of the instrument
    ! reference frame (the same frame as defines the TELCOORD system).
    ! For example, the x or 1st axis of this frame appears in the first row of
    ! AIN_sky. The sky basis is that right-handed frame which has its
    ! x and y axes at declination=0, z axis pointing in the declination
    ! = pi/2 direction, x in the ra=0 and y in the ra=pi/2 (ie, 6 hrs)
    ! directions.
    !
    ! This matrix can be used to convert a vector from the sky basis to
    ! the instrument or TELCOORD basis, as follows:
    !
    !		v_in = AIN_sky * v_sky.
    ! or
    !		v_sky = AIN_sky^-1 * v_in.
    !
    ! Note that the OAL call returns transpose(AIN_sky) = AIN_sky^-1.

    do i = 1, numPositions
      tanTheta = tan(localThetaArcsec(i) * ArcsecToRad)
      detCartesian(1) = -1.0d0
      detCartesian(2) = tanTheta * sin(localPhi(i))
      detCartesian(3) = tanTheta * cos(localPhi(i))

!print *, 'i, detCartesian:', i, detCartesian

      skyCartesian = matmul(transposedInstrumAttitudeMatrix, detCartesian)
      call cartesianToSky(skyCartesian, ra, dec)
      raDeg(i)  = ra  / PiOn180
      decDeg(i) = dec / PiOn180
    end do

    if (present(xDet) .and. present(yDet)) then
      call CAL_releaseMemory(localThetaArcsec)
      call CAL_releaseMemory(localPhi)

    else if (present(thetaArcsec) .and. present(phi)) then
      deallocate(localThetaArcsec, localPhi)
    end if

!**** other CAL_releaseMemory

    if (present(status)) status = localStatus

  end subroutine instToRaDecVector

  !---------------------------------------------------------------------
  subroutine raDecToTanScalar(refRa, refDec, ra, dec, tanX, tanY)!, status)
  !---------------------------------------------------------------------
  ! NOTE that all ras and decs must be in RADIANS!!

    implicit none

    real(double), intent(in)  :: refRa, refDec, ra, dec
    real(double), intent(out) :: tanX, tanY
!    integer, intent(out), optional :: status

    real(double) :: localRa(1), localDec(1), localTanX(1), localTanY(1)
!    integer :: localStatus

    !-------------------------------------------------------------------
    ! Start

    localRa(1) = ra
    localDec(1) = dec

    call raDecToTanVector(refRa, refDec, localRa, localDec&
    , localTanX, localTanY)!, localStatus)

    tanX = localTanX(1)
    tanY = localTanY(1)

!    if (present(status)) status = localStatus

  end subroutine raDecToTanScalar

  !---------------------------------------------------------------------
  subroutine raDecToTanVector(refRa, refDec, ra, dec, tanX, tanY)!, status)
  !---------------------------------------------------------------------
  ! NOTE that all ras and decs must be in RADIANS!!

    implicit none

    real(double), intent(in)            :: refRa,&
                                           refDec,&
                                           ra(:),&
                                           dec(size(ra))
    real(double), intent(out)           :: tanX(size(ra)),&
                                           tanY(size(ra))
!    integer,      intent(out), optional :: status

    real(double) :: srcVector_sky(3), refMatrix_sky(3,3), srcVector_ref(3)
    integer(int32) :: vsize, i
!logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

!    if (present(status)) status = 0

    tanX = 0.0d0 ! default.
    tanY = 0.0d0 ! default.

    vsize = size(ra)

    refMatrix_sky(1, :) = skyToCartesian(refRa, refDec)
    refMatrix_sky(2, :) = skyToCartesian(refRa + PiOverTwo, 0.0d0)
    refMatrix_sky(3, :) = vectorCross(refMatrix_sky(1, :), refMatrix_sky(2, :))

    do i = 1, vsize
      srcVector_sky = skyToCartesian(ra(i), dec(i))
      srcVector_ref = matmul(refMatrix_sky, srcVector_sky)

      tanX(i) = -srcVector_ref(2)
      tanY(i) =  srcVector_ref(3)
    end do

  end subroutine raDecToTanVector

  !---------------------------------------------------------------------
  subroutine tanToRaDecScalar(refRa, refDec, tanX, tanY, ra, dec)!, status)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from a tangent-plane coordinate
  ! system into ra and dec. The x-axis of the tangent plane points to the
  ! right and is in the direction of decreasing ra; the y-axis points up
  ! and is in the direction of increasing dec.
  !
  ! NOTE that all ras and decs must be in RADIANS!!

    implicit none

    real(double), intent(in)  :: refRa, refDec, tanX, tanY
    real(double), intent(out) :: ra, dec
!    integer, intent(out), optional :: status

    real(double) :: tanXvector(1), tanYvector(1), raVector(1), decVector(1)
!    integer(int32) :: foreignStatus

    !-------------------------------------------------------------------
    ! Start

    tanXvector(1) = tanX
    tanYvector(1) = tanY

    call tanToRaDecVector(refRa, refDec, tanXvector, tanYvector&
      , raVector, decVector)!, foreignStatus)

    ra  = raVector(1)
    dec = decVector(1)
!    if (present(status)) status = foreignStatus

  end subroutine tanToRaDecScalar

  !---------------------------------------------------------------------
  subroutine tanToRaDecVector(refRa, refDec, tanX, tanY, ra, dec)!, status)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from a tangent-plane coordinate
  ! system into ra and dec. The x-axis of the tangent plane points to the
  ! right and is in the direction of decreasing ra; the y-axis points up
  ! and is in the direction of increasing dec.
  !
  ! NOTE that all ras and decs must be in RADIANS!!

    implicit none

    real(double), intent(in)            :: refRa,&
                                           refDec,&
                                           tanX(:),&
                                           tanY(size(tanX))
    real(double), intent(out)           :: ra(size(tanX)),&
                                           dec(size(tanX))
!    integer,      intent(out), optional :: status

    real(double) :: srcVector_sky(3), refMatrix_sky(3,3), srcVector_ref(3),&
      norm
    integer(int32) :: vsize, i

    !-------------------------------------------------------------------
    ! Start

!    if (present(status)) status = 0

    ra  = 0.0d0 ! default.
    dec = 0.0d0 ! default.

    vsize = size(tanX)

    refMatrix_sky(1, :) = skyToCartesian(refRa, refDec)
    refMatrix_sky(2, :) = skyToCartesian(refRa + PiOverTwo, 0.0d0)
    refMatrix_sky(3, :) = vectorCross(refMatrix_sky(1, :)&
    , refMatrix_sky(2, :))
!*** how about when ref is pointing straight up or straight down.

    do i = 1, vsize
      norm = sqrt(1.0d0 + tanX(i) * tanX(i) + tanY(i) * tanY(i))
      srcVector_ref(1) = 1.0d0 / norm
      srcVector_ref(2) = -tanX(i) / norm
      srcVector_ref(3) = tanY(i) / norm

      srcVector_sky = matmul(srcVector_ref, refMatrix_sky)
      call cartesianToSky(srcVector_sky, ra(i), dec(i))
    end do

  end subroutine tanToRaDecVector

  !---------------------------------------------------------------------
  subroutine instToTanScalar(tanX, tanY, refRa, refDec, scAttitude&
  , timeStamp, xDet, yDet, thetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from instrument coordimates (ie
  ! either camcoord2 detxy or telcoord theta and phi) into a tangent-
  ! -plane coordinate system. The x-axis of the tangent plane points to
  ! the right and is in the direction of decreasing ra; the y-axis points
  ! up and is in the direction of increasing dec.
  !
  ! NOTE (i) that all ras and decs must be in RADIANS!!
  !
  ! NOTE (ii) that Cal must be set up for the correct instrument by
  ! the calling routine.

    implicit none

    real(double), intent(out)           :: tanX, tanY
    real(double), intent(in)            :: refRa, refDec
    type(SpacecraftAttitudeType), intent(in) :: scAttitude
    ! typedef from caloalutils/src/caloaldefs.f90
    real(double), intent(in)            :: timeStamp
    real(single), intent(in),  optional :: xDet,&
                                           yDet
    real(double), intent(in),  optional :: thetaArcsec,&
                                           phi

    real(single) ::&
      vectorXdet(1), vectorYdet(1)
    real(double) ::&
      vectorTanX(1), vectorTanY(1),&
      vectorThetaArcsec(1), vectorPhi(1)

    !-------------------------------------------------------------------
    ! Start

    tanX = 0.0d0 ! default
    tanY = 0.0d0 ! default

    if (present(xDet) .and. present(yDet)) then
      vectorXdet(1) = xDet
      vectorYdet(1) = yDet

      call instToTanVector(vectorTanX, vectorTanY, refRa, refDec, scAttitude&
        , timeStamp, vectorXdet, vectorYdet)

    else if (present(thetaArcsec) .and. present(phi)) then
      vectorThetaArcsec(1) = thetaArcsec
      vectorPhi(1)         = phi

      call instToTanVector(vectorTanX, vectorTanY, refRa, refDec, scAttitude&
        , timeStamp, thetaArcsec=vectorThetaArcsec, phi=vectorPhi)
    end if

    tanX = vectorTanX(1)
    tanY = vectorTanY(1)

  end subroutine instToTanScalar

  !---------------------------------------------------------------------
  subroutine instToTanVector(tanX, tanY, refRa, refDec, scAttitude&
  , timeStamp, xDet, yDet, thetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from instrument coordimates (ie
  ! either camcoord2 detxy or telcoord theta and phi) into a tangent-
  ! -plane coordinate system. The x-axis of the tangent plane points to
  ! the right and is in the direction of decreasing ra; the y-axis points
  ! up and is in the direction of increasing dec.
  !
  ! NOTE (i) that all ras and decs must be in RADIANS!!
  !
  ! NOTE (ii) that Cal must be set up for the correct instrument by
  ! the calling routine.

    implicit none

    real(double), intent(out)           :: tanX(:),&
                                           tanY(size(tanX))
    real(double), intent(in)            :: refRa, refDec
    type(SpacecraftAttitudeType), intent(in) :: scAttitude
    ! typedef from caloalutils/src/caloaldefs.f90
    real(double), intent(in)            :: timeStamp
    real(single), intent(in),  optional :: xDet(size(tanX)),&
                                           yDet(size(tanX))
    real(double), intent(in),  optional :: thetaArcsec(size(tanX)),&
                                           phi(size(tanX))

    real(double), dimension(:), pointer ::&
      localThetaArcsec, localPhi, cc2x, cc2y, cc2z
    real(double) ::&
      mmToDetUnits, focalLengthMm, tanTheta,&
      srcVector_ins(3), srcVector_ref(3),&
      transpInstrumMatrix_sky(3,3), refMatrix_sky(3,3), refMatrix_ins(3,3)
    integer(int32) ::&
      numPositions, i
    type(SpacecraftAttitudeType) ::&
      instrumAttitude
logical :: testFlag = .false.
integer :: status, ai, bi, ci

    !-------------------------------------------------------------------
    ! Start

    tanX = 0.0d0 ! default
    tanY = 0.0d0 ! default

    numPositions = size(tanX)
ai = numPositions
bi = numPositions
ci = numPositions

    if (present(xDet) .and. present(yDet)) then
      call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
      mmToDetUnits = RadToArcsec / focalLengthMm / detUnitArcsec

!      allocate(cc2X(numPositions), cc2Y(numPositions), cc2Z(numPositions), stat=status)
      allocate(cc2X(ai), cc2Y(bi), cc2Z(ci), stat=status)
if (status .ne. 0) print *, 'alloc status=', status
!allocate(cc2X(numPositions))
!allocate(cc2Y(numPositions))
!allocate(cc2Z(numPositions))
      cc2X = xDet / mmToDetUnits
      cc2Y = yDet / mmToDetUnits
      cc2z = focalLengthMm

      call CAL_camCoord2ToTelCoord(cc2x, cc2y, cc2z, localThetaArcsec, localPhi)

      deallocate(cc2X, cc2Y, cc2z)

    else if (present(thetaArcsec) .and. present(phi)) then
      allocate(localThetaArcsec(numPositions), localPhi(numPositions))
      localThetaArcsec = thetaArcsec
      localPhi = phi
    end if

    instrumAttitude = getInstrumentAttitude(scAttitude, timeStamp)
if (testFlag) print *, 'instrumAttitude%raDeg, instrumAttitude%decDeg: '&
, instrumAttitude%ra / PiOn180, instrumAttitude%dec / PiOn180

    call OAL_toAttitudeMatrix(instrumAttitude, transpInstrumMatrix_sky)
    !
    ! The array returned by this call represents the instrument attitude
    ! matrix AIN_sky which has, in its ith row, a vector of direction
    ! cosines in the sky basis of the ith axis of the instrument
    ! reference frame (the same frame as defines the TELCOORD system).
    ! For example, the x or 1st axis of this frame appears in the first row of
    ! AIN_sky. The sky basis is that right-handed frame which has its
    ! x and y axes at declination=0, z axis pointing in the declination
    ! = pi/2 direction, x in the ra=0 and y in the ra=pi/2 (ie, 6 hrs)
    ! directions.
    !
    ! This matrix can be used to convert a vector from the sky basis to
    ! the instrument or TELCOORD basis, as follows:
    !
    !		v_in = AIN_sky * v_sky.
    ! or
    !		v_sky = AIN_sky^-1 * v_in.
    !
    ! Note that the OAL call in fact returns transpose(AIN_sky) = AIN_sky^-1.

    refMatrix_sky(1, :) = skyToCartesian(refRa, refDec)
    refMatrix_sky(2, :) = skyToCartesian(refRa + PiOverTwo, 0.0d0)
    refMatrix_sky(3, :) = vectorCross(refMatrix_sky(1, :), refMatrix_sky(2, :))

    refMatrix_ins = matmul(refMatrix_sky, transpInstrumMatrix_sky)
!*** the (1,1) number may be inaccurate - should use a better routine.

    do i = 1, numPositions
      tanTheta = tan(localThetaArcsec(i-1) * ArcsecToRad)
      srcVector_ins(1) = -1.0d0
      srcVector_ins(2) = tanTheta * sin(localPhi(i-1))
      srcVector_ins(3) = tanTheta * cos(localPhi(i-1))

      srcVector_ref = matmul(refMatrix_ins, srcVector_ins)

      tanX(i) = -srcVector_ref(2)
      tanY(i) = srcVector_ref(3)
    end do

    if (present(xDet) .and. present(yDet)) then
      call CAL_releaseMemory(localThetaArcsec)
      call CAL_releaseMemory(localPhi)

    else if (present(thetaArcsec) .and. present(phi)) then
      deallocate(localThetaArcsec, localPhi)
    end if

!**** other CAL_releaseMemory

  end subroutine instToTanVector

  !---------------------------------------------------------------------
  subroutine instToTanArray(tanX, tanY, refRa, refDec, scAttitude&
  , timeStamp, xDet, yDet, thetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from instrument coordimates (ie
  ! either camcoord2 detxy or telcoord theta and phi) into a tangent-
  ! -plane coordinate system. The x-axis of the tangent plane points to
  ! the right and is in the direction of decreasing ra; the y-axis points
  ! up and is in the direction of increasing dec.
  !
  ! NOTE (i) that all ras and decs must be in RADIANS!!
  !
  ! NOTE (ii) that Cal must be set up for the correct instrument by
  ! the calling routine.

    implicit none

    real(double), intent(out)           :: tanX(:,:),&
                                           tanY(size(tanX,1),size(tanX,2))
    real(double), intent(in)            :: refRa, refDec
    type(SpacecraftAttitudeType), intent(in) :: scAttitude
    ! typedef from caloalutils/src/caloaldefs.f90
    real(double), intent(in)            :: timeStamp
    real(single), intent(in),  optional :: xDet(size(tanX,1),size(tanX,2)),&
                                           yDet(size(tanX,1),size(tanX,2))
    real(double), intent(in),  optional :: thetaArcsec(size(tanX,1)&
                                                      ,size(tanX,2)),&
                                           phi(size(tanX,1),size(tanX,2))

    integer(int32) :: arrayShape(2), vectorShape(1)
    real(single) ::&
      vectorXdet(size(tanX,1)*size(tanX,2)),&
      vectorYdet(size(tanX,1)*size(tanX,2))
    real(double) ::&
      vectorTanX(size(tanX,1)*size(tanX,2)),&
      vectorTanY(size(tanX,1)*size(tanX,2)),&
      vectorThetaArcsec(size(tanX,1)*size(tanX,2)),&
      vectorPhi(size(tanX,1)*size(tanX,2))

    !-------------------------------------------------------------------
    ! Start

    tanX = 0.0d0 ! default
    tanY = 0.0d0 ! default

    arrayShape(1) = size(tanX, 1)
    arrayShape(2) = size(tanX, 2)
    vectorShape(1) = arrayShape(1) * arrayShape(2)

    if (present(xDet) .and. present(yDet)) then
      vectorXdet = reshape(xDet, vectorShape)
      vectorYdet = reshape(yDet, vectorShape)

      call instToTanVector(vectorTanX, vectorTanY, refRa, refDec, scAttitude&
        , timeStamp, vectorXdet, vectorYdet)

    else if (present(thetaArcsec) .and. present(phi)) then
      vectorThetaArcsec = reshape(thetaArcsec, vectorShape)
      vectorPhi         = reshape(phi,         vectorShape)

      call instToTanVector(vectorTanX, vectorTanY, refRa, refDec, scAttitude&
        , timeStamp, thetaArcsec=vectorThetaArcsec, phi=vectorPhi)
    end if

    tanX = reshape(vectorTanX, arrayShape)
    tanY = reshape(vectorTanY, arrayShape)

  end subroutine instToTanArray

  !---------------------------------------------------------------------
  subroutine tanToInstScalar(tanX, tanY, refRa, refDec, scAttitude&
  , timeStamp, xDet, yDet, thetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from a tangent-plane coordinate
  ! system into instrument coordimates (ie either camcoord2 detxy or
  ! telcoord theta and phi). The x-axis of the tangent plane points to
  ! the right and is in the direction of decreasing ra; the y-axis points
  ! up and is in the direction of increasing dec.
  !
  ! NOTE (i) that all ras and decs must be in RADIANS!!
  !
  ! NOTE (ii) that Cal must be set up for the correct instrument by
  ! the calling routine.

    implicit none

    real(double),                 intent(in)  :: refRa,&
                                                 refDec,&
                                                 tanX,&
                                                 tanY
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    ! typedef from caloalutils/src/caloaldefs.f90
    real(double),                 intent(in)  :: timeStamp
    real(single),      optional,  intent(out) :: xDet,&
                                                 yDet
    real(double),      optional,  intent(out) :: thetaArcsec,&
                                                 phi

    real(single) ::&
      vectorXdet(1), vectorYdet(1)
    real(double) ::&
      vectorTanX(1), vectorTanY(1),&
      vectorThetaArcsec(1), vectorPhi(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(xDet)) xDet = 0.0
    if (present(yDet)) yDet = 0.0
    if (present(thetaArcsec)) thetaArcsec = 0.0d0
    if (present(phi)) phi = 0.0d0

    vectorTanX(1) = tanX
    vectorTanY(1) = tanY

    if (present(xDet) .and. present(yDet)) then
      call tanToInstVector(vectorTanX, vectorTanY, refRa, refDec, scAttitude&
        , timeStamp, vectorXdet, vectorYdet)

      xDet = vectorXdet(1)
      yDet = vectorYdet(1)

    else if (present(thetaArcsec) .and. present(phi)) then

      call tanToInstVector(vectorTanX, vectorTanY, refRa, refDec, scAttitude&
        , timeStamp, thetaArcsec=vectorThetaArcsec, phi=vectorPhi)

      thetaArcsec = vectorThetaArcsec(1)
      phi         = vectorPhi(1)
    end if

  end subroutine tanToInstScalar

  !---------------------------------------------------------------------
  subroutine tanToInstVector(tanX, tanY, refRa, refDec, scAttitude&
  , timeStamp, xDet, yDet, thetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from a tangent-plane coordinate
  ! system into instrument coordimates (ie either camcoord2 detxy or
  ! telcoord theta and phi). The x-axis of the tangent plane points to
  ! the right and is in the direction of decreasing ra; the y-axis points
  ! up and is in the direction of increasing dec.
  !
  ! NOTE (i) that all ras and decs must be in RADIANS!!
  !
  ! NOTE (ii) that Cal must be set up for the correct instrument by
  ! the calling routine.

    implicit none

    real(double),                 intent(in)  :: refRa,&
                                                 refDec,&
                                                 tanX(:),&
                                                 tanY(size(tanX))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    ! typedef from caloalutils/src/caloaldefs.f90
    real(double),                 intent(in)  :: timeStamp
    real(single),      optional,  intent(out) :: xDet(size(tanX)),&
                                                 yDet(size(tanX))
    real(double),      optional,  intent(out) :: thetaArcsec(size(tanX)),&
                                                 phi(size(tanX))

    real(double) :: srcVector_sky(3), refMatrix_sky(3,3), srcVector_ref(3),&
      norm, mmToDetUnits, focalLengthMm
    integer(int32) :: numPositions, i
    type(BoresightT) ::&
      boresight ! typedef from cal/f90/caltypes.f90
    type(EquatorialDirectionType) ::&
      skyVectors(size(tanX)) ! typedef from caloalutils/src/caloaldefs.f90
    real(double), dimension(:), pointer ::&
      thetaArcsecArg, phiArg, cc2x, cc2y, cc2z

    !-------------------------------------------------------------------
    ! Start

    if (present(xDet)) xDet = 0.0
    if (present(yDet)) yDet = 0.0
    if (present(thetaArcsec)) thetaArcsec = 0.0d0
    if (present(phi)) phi = 0.0d0

    numPositions = size(tanX)

    refMatrix_sky(1, :) = skyToCartesian(refRa, refDec)
    refMatrix_sky(2, :) = skyToCartesian(refRa + PiOverTwo, 0.0d0)
    refMatrix_sky(3, :) = vectorCross(refMatrix_sky(1, :)&
    , refMatrix_sky(2, :))
!*** how about when ref is pointing straight up or straight down.

    do i = 1, numPositions
      norm = sqrt(1.0d0 + tanX(i) * tanX(i) + tanY(i) * tanY(i))
      srcVector_ref(1) = 1.0d0 / norm
      srcVector_ref(2) = -tanX(i) / norm
      srcVector_ref(3) = tanY(i) / norm

      srcVector_sky = matmul(srcVector_ref, refMatrix_sky)
      call cartesianToSky(srcVector_sky, skyVectors(i)%ra, skyVectors(i)%dec)
    end do

    boresight = CAL_getBoresight(timeStamp)
    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
    mmToDetUnits = RadToArcsec / focalLengthMm / detUnitArcsec
    call CAL_withSacCoordOrientation(boresight, scAttitude)
    call CAL_projectOntoTelCoord(boresight, skyVectors, thetaArcsecArg, phiArg) 
    ! NOTE! Units of phiArg are radians.
    call CAL_telCoordToCamCoord2(thetaArcsecArg, phiArg, cc2x, cc2y, cc2z)

    if (present(xDet)) xDet = cc2X * mmToDetUnits
    if (present(yDet)) yDet = cc2Y * mmToDetUnits
    if (present(thetaArcsec)) thetaArcsec = thetaArcsecArg
    if (present(phi)) phi = phiArg

    call CAL_releaseMemory(thetaArcsecArg)
    call CAL_releaseMemory(phiArg)
    call CAL_releaseMemory(cc2x, cc2y, cc2z)

  end subroutine tanToInstVector

  !---------------------------------------------------------------------
  subroutine tanToInstArray(tanX, tanY, refRa, refDec, scAttitude&
  , timeStamp, xDet, yDet, thetaArcsec, phi)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from a tangent-plane coordinate
  ! system into instrument coordimates (ie either camcoord2 detxy or
  ! telcoord theta and phi). The x-axis of the tangent plane points to
  ! the right and is in the direction of decreasing ra; the y-axis points
  ! up and is in the direction of increasing dec.
  !
  ! NOTE (i) that all ras and decs must be in RADIANS!!
  !
  ! NOTE (ii) that Cal must be set up for the correct instrument by
  ! the calling routine.

    implicit none

    real(double),               intent(in)  :: refRa,&
                                               refDec,&
                                               tanX(:,:),&
                                               tanY(size(tanX,1),size(tanX,2))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    ! typedef from caloalutils/src/caloaldefs.f90
    real(double),               intent(in)  :: timeStamp
    real(single),    optional,  intent(out) :: xDet(size(tanX,1),size(tanX,2)),&
                                               yDet(size(tanX,1),size(tanX,2))
    real(double),    optional,  intent(out) :: thetaArcsec(size(tanX,1)&
                                                         , size(tanX,2)),&
                                               phi(size(tanX,1),size(tanX,2))

    integer(int32) :: arrayShape(2), vectorShape(1)
    real(single) ::&
      vectorXdet(size(tanX,1)*size(tanX,2)),&
      vectorYdet(size(tanX,1)*size(tanX,2))
    real(double) ::&
      vectorTanX(size(tanX,1)*size(tanX,2)),&
      vectorTanY(size(tanX,1)*size(tanX,2)),&
      vectorThetaArcsec(size(tanX,1)*size(tanX,2)),&
      vectorPhi(size(tanX,1)*size(tanX,2))

    !-------------------------------------------------------------------
    ! Start

    if (present(xDet)) xDet = 0.0
    if (present(yDet)) yDet = 0.0
    if (present(thetaArcsec)) thetaArcsec = 0.0d0
    if (present(phi)) phi = 0.0d0

    arrayShape(1) = size(tanX, 1)
    arrayShape(2) = size(tanX, 2)
    vectorShape(1) = arrayShape(1) * arrayShape(2)
    vectorTanX = reshape(tanX, vectorShape)
    vectorTanY = reshape(tanY, vectorShape)

    if (present(xDet) .and. present(yDet)) then
      call tanToInstVector(vectorTanX, vectorTanY, refRa, refDec, scAttitude&
        , timeStamp, vectorXdet, vectorYdet)

      xDet = reshape(vectorXdet, arrayShape)
      yDet = reshape(vectorYdet, arrayShape)

    else if (present(thetaArcsec) .and. present(phi)) then

      call tanToInstVector(vectorTanX, vectorTanY, refRa, refDec, scAttitude&
        , timeStamp, thetaArcsec=vectorThetaArcsec, phi=vectorPhi)

      thetaArcsec = reshape(vectorThetaArcsec, arrayShape)
      phi         = reshape(vectorPhi,         arrayShape)
    end if

  end subroutine tanToInstArray

  !---------------------------------------------------------------------
  function skyToCartesian(ra, dec) result(vector)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: ra, dec 
    real(double)             :: vector(3)

    real (double) :: cosDec 

    !-------------------------------------------------------------------
    ! Start

    cosDec = cos(dec)
    vector(1) = cos(ra) * cosDec
    vector(2) = sin(ra) * cosDec
    vector(3) = sin(dec)

  end function skyToCartesian

  !---------------------------------------------------------------------
  subroutine cartesianToSky(vector, ra, dec)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: vector(3) 
    real(double), intent(out) :: ra, dec

    !-------------------------------------------------------------------
    ! Start

    if (vector(1) .ne. 0.0 .and. vector(2) .ne. 0.0) then
      ra = atan2(vector(2), vector(1))
      do while(ra .lt. 0.0)
        ra = ra + TwoPi
      end do
    else
      ra = 0.0d0
    end if
    dec = asin(vector(3))

  end subroutine cartesianToSky

  !---------------------------------------------------------------------
  subroutine getThetaPhiMaps(wcs, scAttitude, timeStamp, thetaMap, phiMap)
  !---------------------------------------------------------------------
  ! NOTE! The cal must be set up for the correct instrument before this
  ! is called.

    implicit none

    type(WcsT),                   intent(in)  :: wcs
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    real(double),                 intent(in)  :: timeStamp
    real(double),                 intent(out) :: thetaMap(:,:),&
                                                 phiMap(size(thetaMap,1)&
                                                      , size(thetaMap,2))

    integer(int16) :: xSize, ySize
    integer(int16), pointer :: xPixels(:), yPixels(:)
    real(double), allocatable :: tanX(:), tanY(:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(thetaMap, 1)
    ySize = size(thetaMap, 2)
    xPixels => indicize(xSize) ! in ssc_misc
    yPixels => indicize(ySize) ! in ssc_misc
    allocate(tanX(xSize), tanY(ySize))
    tanX = (xPixels - wcs%x%refPixels) * abs(wcs%x%pixelSizeWorld) * PiOn180
    tanY = (yPixels - wcs%y%refPixels) * abs(wcs%y%pixelSizeWorld) * PiOn180

    call tanToInst(spread(tanX, 2, ySize), spread(tanY, 1, xSize)&
      , wcs%x%refWorld, wcs%y%refWorld, scAttitude&
      , timeStamp, thetaArcsec=thetaMap, phi=phiMap)

    deallocate(xPixels, yPixels)

  end subroutine getThetaPhiMaps

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module coordinate_old
